from typing import Any

import pandas as pd
import pydantic

from mlstac.specification.collection.datamodel import Collection


class MLSTAC(pydantic.BaseModel):
    collection: Collection
    metadata: Any

    @pydantic.field_validator("metadata")
    def check_metadata(cls, v):
        if not isinstance(v, pd.DataFrame):
            raise ValueError("The metadata must be a pandas DataFrame")
        return v

    def __str__(self):
        return f"MLSTAC(collection=<dict>, metadata=<pandas.DataFrame>)"

    def __repr__(self):
        return self.__str__()



import pathlib
from typing import Union

from mdutils.mdutils import MdUtils


def add_documentation(ml_collection, outfile: Union[str, pathlib.Path]):
    if isinstance(outfile, str):
        outfile = pathlib.Path(outfile)
    outfile.parent.mkdir(parents=True, exist_ok=True)

    # Create the README.md file
    mdFile = MdUtils(file_name=outfile)

    # Create yaml header ------------------------------------
    mdFile.new_line("---")
    mdFile.new_line(f"license: {ml_collection.license}")
    mdFile.new_line(f"task_categories:")
    for item in ml_collection.ml_task:
        mdFile.new_line(f"  - {item}")
    mdFile.new_line(f"language:")
    mdFile.new_line(f"- en")
    if ml_collection.keywords is not None:
        mdFile.new_line(f"tags:")
        for tag in ml_collection.keywords:
            mdFile.new_line(f"  - {tag}")
    mdFile.new_line(f"pretty_name: {ml_collection.id}")
    mdFile.new_line("---")

    # Set title, subtitle and description --------------------
    mdFile.new_header(level=1, title=f"{ml_collection.id}")
    if ml_collection.title is not None:
        mdFile.new_line(f"{ml_collection.title}", bold_italics_code="cib")
    mdFile.new_paragraph(f"{ml_collection.description}\n")
    mdFile.new_paragraph(f"ML-STAC Snippet", bold_italics_code="b")

    # Add code snippet ---------------------------------------
    mdFile.new_line("```python")
    mdFile.new_line(f"import mlstac")
    mdFile.new_line(f"dataset = mlstac.load('...')")
    mdFile.new_line("```")

    # Add data raw repository --------------------------------
    if ml_collection.ml_spectral is not None:
        mdFile.new_paragraph(
            "Sensor: " + ml_collection.ml_spectral.sensor, bold_italics_code="b"
        )

        mdFile.new_paragraph(
            "ML-STAC Task: " + ", ".join(ml_collection.ml_task), bold_italics_code="b"
        )

    if ml_collection.ml_raw_data_url is not None:
        mdFile.new_paragraph(
            f"Data raw repository:  "
            + mdFile.new_inline_link(link=ml_collection.ml_raw_data_url),
            bold_italics_code="b",
        )

    if ml_collection.ml_discussion_url is not None:
        mdFile.new_paragraph(
            "Dataset discussion:  "
            + mdFile.new_inline_link(link=ml_collection.ml_discussion_url),
            bold_italics_code="b",
        )

    if ml_collection.ml_reviewers is not None:
        mdFile.new_paragraph(
            f"Review mean score:  {ml_collection.get_review_mean_score()}",
            bold_italics_code="b",
        )

    if ml_collection.ml_split_strategy is not None:
        mdFile.new_paragraph(
            f"Split_strategy:  " + ml_collection.ml_split_strategy,
            bold_italics_code="b",
        )

    if ml_collection.ml_paper is not None:
        mdFile.new_paragraph(
            f"Paper:  " + mdFile.new_inline_link(link=ml_collection.ml_paper),
            bold_italics_code="b",
        )

    # Add Provider --------------------------------------------
    mdFile.new_header(level=2, title=f"Data Providers")
    list_of_strings = ["Name", "Role", "URL"]
    for p in ml_collection.providers.field:
        list_of_strings.extend([p.name, p.roles, p.url])
    mdFile.new_table(
        columns=3,
        rows=len(ml_collection.providers.field) + 1,
        text=list_of_strings,
        text_align="center",
    )

    # Add Authors ---------------------------------------------
    # if len(self.ml_authors.authors) > 0:
    #    mdFile.new_header(level=2, title=f"Authors")
    #    list_of_strings = ["Name", "Organization"]
    #    for p in self.ml_authors:
    #        list_of_strings.extend([p['name'], p['organization']])
    # mdFile.new_table(columns=2, rows=len(self.ml_authors.authors)+1, text=list_of_strings, text_align='center')

    # Add Curators --------------------------------------------
    if ml_collection.ml_curators is not None:
        mdFile.new_header(level=2, title=f"Curators")
        list_of_strings = ["Name", "Organization", "URL"]
        for p in ml_collection.ml_curators.curators:
            list_of_strings.extend([p.name, p.organization, p.links[0].href])
        mdFile.new_table(
            columns=3,
            rows=len(ml_collection.ml_curators.curators) + 1,
            text=list_of_strings,
            text_align="center",
        )

    # Add Reviewers --------------------------------------------
    if ml_collection.ml_reviewers is not None:
        mdFile.new_header(level=2, title=f"Reviewers")
        list_of_strings = ["Name", "Organization", "URL", "Score"]
        for reviewer in ml_collection.ml_reviewers.reviewers:
            p = reviewer.reviewer
            s = reviewer.score
            list_of_strings.extend([p.name, p.organization, p.links[0].href, s])
        mdFile.new_table(
            columns=4,
            rows=len(ml_collection.ml_reviewers.reviewers) + 1,
            text=list_of_strings,
            text_align="center",
        )

    # Add Labels ----------------------------------------------
    if ml_collection.ml_target is not None:
        mdFile.new_header(level=2, title="Labels")
        list_of_strings = ["Name", "Value"]
        for k, v in ml_collection.ml_target.labels.items():
            list_of_strings.extend([k, v])
        mdFile.new_table(
            columns=2,
            rows=len(ml_collection.ml_target.labels) + 1,
            text=list_of_strings,
            text_align="center",
        )

        # Add Dimensions ------------------------------------------
        if ml_collection.ml_dimensions is not None:
            mdFile.new_header(level=2, title=f"Dimensions")
            for k, v in ml_collection.ml_dimensions.model_dump().items():
                if len(v) > 0:
                    mdFile.new_header(level=3, title=f"{k}")
                    list_of_strings = ["Axis", "Name", "Description"]
                    for p in v:
                        list_of_strings.extend([p["axis"], p["name"], p["description"]])
                    mdFile.new_table(
                        columns=3,
                        rows=len(v) + 1,
                        text=list_of_strings,
                        text_align="center",
                    )

        # Add Spectral Bands --------------------------------------
        if ml_collection.ml_spectral is not None:
            mdFile.new_header(level=2, title=f"Spectral Bands")
            list_of_strings = [
                "Name",
                "Common Name",
                "Description",
                "Center Wavelength",
                "Full Width Half Max",
                "Index",
            ]
            for p in ml_collection.ml_spectral.bands:
                list_of_strings.extend(
                    [
                        p.name,
                        p.common_name,
                        p.description,
                        p.center_wavelength,
                        p.full_width_half_max,
                        p.index,
                    ]
                )
            mdFile.new_table(
                columns=6,
                rows=len(ml_collection.ml_spectral.bands) + 1,
                text=list_of_strings,
                text_align="center",
            )

        file = mdFile.get_md_text().replace("\n\n\n  \n", "").replace("  \n", "\n")
        with open(outfile, "w") as f:
            f.write(file)



from typing import List, Literal, Optional

import pydantic

from mlstac.specification.collection.extension_contacts import Contact


class Reviewer(pydantic.BaseModel):
    """This extension provides a way to define reviewers. Useful for
    measuring the quality of a dataset.

    Fields:
        name (str): Name of the reviewer.
        score (Literal[0, 1, 2, 3, 4, 5]): Score of the reviewer.
            The score must be between 0 and 5. Higher is better.
        url (str): A public url where to find more information about
            the review process. Usually is a link to a github issue.
    """

    reviewer: Contact
    score: Optional[Literal[0, 1, 2, 3, 4, 5]] = None


class Reviewers(pydantic.BaseModel):
    """This extension define a review process. Useful for
    measuring the quality of a dataset.
    """

    reviewers: Optional[List[Reviewer]] = None



from typing import List, Optional

import pydantic

from mlstac.specification.collection.extension_contacts import Contact


class Curators(pydantic.BaseModel):
    """This extension provides a way to define curators. Useful for
    measuring the quality of a dataset.

    Fields:
        curators (List[Contact]): A list of curators.
    """

    curators: Optional[List[Contact]] = None



import json
import pathlib
from typing import List, Literal, Optional, Union

import pydantic

import mlstac.specification.catalog.datamodel as mlstac_catalog
import mlstac.specification.collection.extension_authors as mlstac_authors
import mlstac.specification.collection.extension_auto_documentation as auto_documentation
import mlstac.specification.collection.extension_curators as mlstac_curators
import mlstac.specification.collection.extension_dimensions as mlstac_dimensions
import mlstac.specification.collection.extension_reviewers as mlstac_reviewers
import mlstac.specification.collection.extension_sensors as mlstac_sensors
import mlstac.specification.collection.extension_spectralbands as mlstac_spectralbands
import mlstac.specification.collection.extension_target as mlstac_target
import mlstac.specification.collection.stac as stac


class Collection(stac.STACCollection):

    # Required ML-STAC fields -------------------------------
    mlstac_version: str = "0.1.0"
    ml_task: List[str]
    ml_catalog: mlstac_catalog.Catalog
    ml_dataset_version: str


    # Optional ML-STAC fields -------------------------------
    ml_target: Optional[mlstac_target.Target] = None
    ml_authors: Optional[mlstac_authors.Authors] = None
    ml_curators: Optional[mlstac_curators.Curators] = None
    ml_reviewers: Optional[mlstac_reviewers.Reviewers] = None
    ml_dimensions: Optional[mlstac_dimensions.TensorDimensions] = None
    ml_spectral: Optional[mlstac_spectralbands.SpectralBands] = None
    ml_split_strategy: Optional[
        Literal["random", "stratified", "systematic", "other"]
    ] = None
    ml_raw_data_url: Optional[str] = None
    ml_discussion_url: Optional[str] = None
    ml_paper: Optional[str] = None

    def add_extension(self, extension: str) -> None:
        """Add an STAC extension to the ML-STAC collection."""

        # add extension
        if self.stac_extensions is None:
            self.stac_extensions = [extension]
        else:
            self.stac_extensions.append(extension)

        # remove duplicates
        self.stac_extensions = list(set(self.stac_extensions))

        return None

    def add_author(self, **kwargs: dict) -> None:
        """This method adds the authors of the dataset."""

        # Get the previous authors
        if self.ml_authors is not None:
            dataset_authors = self.ml_authors.authors
        else:
            dataset_authors = []

        # Check if the author is already defined
        if any([d.name == kwargs["name"] for d in dataset_authors]):
            for d in dataset_authors:
                if d.name == kwargs["name"]:
                    dataset_authors.remove(d)

        # Load and append the new author
        new_contact = mlstac_curators.Contact(**kwargs)
        dataset_authors.append(new_contact)

        # Save the new authors
        self.ml_authors = mlstac_authors.Authors(authors=dataset_authors)

        # add contact stac extension
        self.add_extension(
            "https://stac-extensions.github.io/contacts/v0.1.1/schema.json"
        )

        print("Author added to self.ml_authors")

        return None

    def add_curator(self, **kwargs: dict) -> None:
        """This method adds the curators of the dataset.
        Useful for reporting errors or requesting changes.
        """

        # Get the previous curators
        if self.ml_curators is not None:
            dataset_curators = self.ml_curators.curators
        else:
            dataset_curators = []

        # Check if the curator is already defined
        if any([d.name == kwargs["name"] for d in dataset_curators]):
            for d in dataset_curators:
                if d.name == kwargs["name"]:
                    dataset_curators.remove(d)

        # Load and append the new curator
        new_contact = mlstac_curators.Contact(**kwargs)
        dataset_curators.append(new_contact)

        # Save the new curators
        self.ml_curators = mlstac_curators.Curators(curators=dataset_curators)

        # add contact stac extension if not already added
        self.add_extension(
            "https://stac-extensions.github.io/contacts/v0.1.1/schema.json"
        )

        print("Curator added to self.ml_curators")
        return None

    def add_labels(self, **kwargs: dict) -> None:
        """This method adds the labels of the dataset.
        Useful for TensorClassification, ObjectDetection and
        SemanticSegmentation tasks.
        """
        self.ml_target = mlstac_target.Target(**kwargs)
        print("Labels added to self.ml_target")
        return None

    def add_dimension(self, **kwargs: dict) -> None:

        # Get the previous dimensions
        if self.ml_dimensions is not None:
            tensor_dimensions = self.ml_dimensions
        else:
            tensor_dimensions = mlstac_dimensions.TensorDimensions(dimensions=[])

        init_dimensions = getattr(tensor_dimensions, "dimensions")

        # Check if the axis is already defined
        try:
            if any([d.axis == kwargs["axis"] for d in init_dimensions.dimensions]):
                for d in init_dimensions.dimensions:
                    if d.axis == kwargs["axis"]:
                        init_dimensions.dimensions.remove(d)
        except:
            pass

        # Load and append the new curator
        dimension = mlstac_dimensions.TensorDimension(**kwargs)
        init_dimensions.append(dimension)

        # Save the new dimensions
        setattr(tensor_dimensions, "dimensions", init_dimensions)

        self.ml_dimensions = tensor_dimensions
        print("Dimensions added to self.ml_dimensions")

        return None

    def add_spectral_band(self, **kwargs: dict) -> None:
        """Add a spectral band to the dataset."""

        # Get the previous spectral bands
        if self.ml_spectral is not None:
            init_bands = self.ml_spectral.bands
        else:
            init_bands = []

        # Check if the axis is already defined
        if any([d.name == kwargs["name"] for d in init_bands]):
            for d in init_bands:
                if d.name == kwargs["name"]:
                    init_bands.remove(d)

        # Load and append the new curator
        band = mlstac_spectralbands.SpectralBand(**kwargs)
        init_bands.append(band)

        # Save the new dimensions
        self.ml_spectral = mlstac_spectralbands.SpectralBands(bands=init_bands)

        print("Dimensions added to self.ml_spectral")

        return None

    def add_sensor(self, sensor: str):
        """Add the sensor of the spectral bands."""
        if self.ml_spectral is None:
            raise ValueError("There are no spectral bands")

        self.ml_spectral = mlstac_spectralbands.SpectralBands(
            bands=self.ml_spectral.bands, axis=self.ml_spectral.axis, sensor=sensor
        )

        print("Specral information added to self.ml_spectral")
        return None

    def add_sentinel2(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_sentinel2(self, bands=bands)
        return None

    def add_landsat1(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat1(self, bands=bands)
        return None

    def add_landsat2(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat2(self, bands=bands)
        return None

    def add_landsat3(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat3(self, bands=bands)
        return None

    def add_landsat4_mss(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat4_mss(self, bands=bands)
        return None

    def add_landsat4_tm(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat4_tm(self, bands=bands)
        return None

    def add_landsat5_mss(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat5_mss(self, bands=bands)
        return None

    def add_landsat5_tm(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat5_tm(self, bands=bands)
        return None

    def add_landsat7(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat7(self, bands=bands)
        return None

    def add_landsat8(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat8(self, bands=bands)
        return None

    def add_landsat9(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_landsat9(self, bands=bands)
        return None

    def add_eo1_ali(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_eo1_ali(self, bands=bands)
        return None

    def add_aster(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_aster(self, bands=bands)
        return None

    def add_modis(self, bands: str = "all"):
        self.ml_spectral = None
        mlstac_sensors.add_modis(self, bands=bands)
        return None

    def add_raw_data_url(self, url: str):
        self.ml_raw_data_url = str(pydantic.AnyHttpUrl(url))
        print("Raw information added to self.ml_raw_data_url")
        return None

    def add_discussion_url(self, url: str):
        self.ml_discussion_url = str(pydantic.AnyHttpUrl(url))
        print("Discussion added to self.ml_discuss_url")
        return None

    def add_split_strategy(self, split_strategy: str):
        self.ml_split_strategy = split_strategy
        print("Split added to self.ml_split")
        return None

    def add_paper(self, url: str):
        """This method adds the paper of the dataset."""
        self.ml_paper = str(pydantic.AnyHttpUrl(url))
        print("Paper added to self.ml_discussion_url")
        return None

    def create_markdown(self, outfile: Union[str, pathlib.Path]) -> None:
        """Add the documentation of the MLSTAC collection."""
        auto_documentation.add_documentation(ml_collection=self, outfile=outfile)
        return None

    def create_json(self, outfile: Union[str, pathlib.Path]) -> None:
        """Create the ML-STAC file."""
        json_dict = self.model_dump()
        with open(outfile, "w") as f:
            json.dump(json_dict, f, indent=4, default=str)



import base64
import json
import pathlib
import re
import zlib
from typing import List, Union
from urllib.parse import urlparse

import pandas as pd
import requests


def is_valid_url(url: str) -> bool:
    """Check if a URL is valid

    Args:
        url (str): The URL to check.

    Returns:
        bool: True if the URL is valid, False otherwise.
    """
    try:
        result = urlparse(url)
        return all([result.scheme in ["http", "https"], result.netloc])
    except ValueError:
        return False


def compress_string(input_string):
    # Compress the string using zlib
    compressed_data = zlib.compress(input_string.encode("utf-8"))

    # Encode the compressed data to base64 to get a string output
    compressed_string = base64.b64encode(compressed_data).decode("utf-8")

    return compressed_string


def from_mlstac_to_df_download(
    metadata: dict, path: Union[pathlib.Path, str], splits: List[str]
) -> pd.DataFrame:
    # Load the metadata
    concat_list = []
    for split_idx in splits:

        # Save the metadata
        metadata_url_file = metadata["ml_catalog"][split_idx]["metadata_file"]
        metadata_checksum = metadata["ml_catalog"][split_idx]["metadata_checksum"]
        metadata_local_file = (
            pathlib.Path(path) / split_idx / "mls" / metadata_url_file.split("/")[-1]
        )
        metadata_local_file.parent.mkdir(parents=True, exist_ok=True)
        concat_list.append(
            {
                "url": metadata_url_file,
                "local": metadata_local_file,
                "checksum": metadata_checksum,
            }
        )

        # Save the data files
        data_files = metadata["ml_catalog"][split_idx]["data_files"]
        for idx, data_file in enumerate(data_files):
            data_url_file = data_file
            data_local_file = (
                pathlib.Path(path) / split_idx / "mls" / data_file.split("/")[-1]
            )
            data_checksum = metadata["ml_catalog"][split_idx]["data_checksum"][idx]
            concat_list.append(
                {
                    "url": data_url_file,
                    "local": data_local_file,
                    "checksum": data_checksum,
                }
            )

    # Download the files
    download_db = pd.DataFrame(concat_list)

    return download_db


def load_mlstac_collection_file(snippet: Union[str, pathlib.Path, dict]) -> dict:
    """Load a MLSTAC collection file

    Args:
        snippet (Union[str, pathlib.Path, dict]): The snippet of the

    Returns:
        dict: The MLSTAC collection file.
    """

    # Load the MLSTAC collection file
    if not (isinstance(snippet, str) or isinstance(snippet, pathlib.Path)):
        return snippet, "remote"
    else:
        snippet = str(snippet)

    # Get the snippet file name if is a URL
    regex_compile = re.compile(r"^[\w-]+/[\w-]+$")

    # is your dataset in HF Datasets?
    if regex_compile.match(snippet):
        snippet_file = (
            f"https://huggingface.co/datasets/{snippet}/resolve/main/main.json"
        )
        with requests.get(snippet_file) as response:
            mlstac_collection = response.json()
        return mlstac_collection, "remote"

    # is your dataset a local file?
    if pathlib.Path(snippet).exists():
        snippet_file = snippet
        with open(snippet_file) as f:
            mlstac_collection = json.load(f)
        return mlstac_collection, "local"

    # is your dataset a valid URL?
    if is_valid_url(snippet):
        snippet_file = snippet
        with requests.get(snippet_file) as response:
            mlstac_collection = response.json()
        return mlstac_collection, "remote"


def download_file_with_recovery(
    url: str,
    local_filename: pathlib.Path,
    checksum: int,
    chunk_size: int = 1024 * 1024 * 100,
    quiet: bool = False,
) -> None:
    """Download a file with recovery capabilities

    This function downloads a file from a URL with the capability of
    resuming the download from a specific byte. If the file is already
    downloaded, the function checks the checksum and returns the
    local filename.

    Args:
        url (str): The URL of the file to download.
        local_filename (pathlib.Path): The local filename to save the file.
        checksum (int): The checksum of the file.
        chunk_size (int, optional): The size of the chunks to download. By
            default, it is 100 MB.
        quiet (bool, optional): If True, the function does not print any
            message. By default, it is False.

    Returns:
        None: The function does not return any value.
    """

    # Determine the starting point for the download
    start = 0
    if pathlib.Path(local_filename).exists():
        start = pathlib.Path(local_filename).stat().st_size

    # Check if the file is already downloaded
    if start == checksum:
        return None

    # Define the headers for the HTTP request
    headers = {"Range": f"bytes={start}-"}

    # Download the file
    with requests.get(url, headers=headers, stream=True) as response:

        # Check if the response is successful
        response.raise_for_status()

        # Open the file in append mode if the file already exists
        mode = "ab" if start > 0 else "wb"

        with open(local_filename, mode) as f:

            # Print a message if the download is resumed
            if start != 0:
                start_gb = start / 1024 / 1024 / 1024
                if not quiet:
                    print(f"Resuming download from byte {start} ({start_gb:.2f} GB)")

            # Download the file
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:
                    f.write(chunk)

    return None


def from_mlstac_to_metadata_parquet(
    mlstac_collection: dict, state: str, snippet: pathlib.Path
) -> pd.DataFrame:
    """Convert a MLSTAC collection to a metadata parquet file

    Every MLSTAC collection has three splits: train, validation, and test.
    This function reads the metadata of each split and concatenates them
    into a single DataFrame. This is useful to create a metadata DataFrame
    that can be queried to extract information about the dataset.

    Args:
        mlstac_collection (dict): The MLSTAC collection.
        state (str): The state of the MLSTAC collection. It can be
            "remote" or "local".
        snippet (pathlib.Path): The snippet of the MLSTAC collection.

    Returns:
        pd.DataFrame: The metadata as a DataFrame.
    """

    splits = ["train", "validation", "test"]
    concat_list = []
    if state == "remote":
        for split_idx in splits:
            # Get the metadata URL
            metadata_url_file = mlstac_collection["ml_catalog"][split_idx][
                "metadata_file"
            ]
            if metadata_url_file is None:
                continue

            # Download the metadata
            metadata_dataset = pd.read_parquet(str(metadata_url_file))
            concat_list.append(metadata_dataset)
    elif state == "local":
        data_path: pathlib.Path = pathlib.Path(snippet).parent
        for split_idx in splits:
            # Get the metadata file
            metadata_local_file = data_path / split_idx / "mls" / "metadata.parquet"
            if metadata_local_file.exists():
                metadata_dataset = pd.read_parquet(metadata_local_file)
                metadata_dataset["file_path"] = list(
                    metadata_dataset["url"].apply(
                        lambda x: metadata_local_file.parent / x.split("/")[-1]
                    )
                )
                concat_list.append(metadata_dataset)

    results = pd.concat(concat_list)
    results["state"] = state

    return results



import json
import pathlib
from typing import List, Union

import tqdm

import mlstac.specification.collection.stac as stac
import mlstac.core

from mlstac.main import download, get_data, load, prepare_data
from mlstac.specification.catalog.datamodel import Catalog, SingleCatalog
from mlstac.specification.collection.datamodel import Collection


def create_file(
    files: List[str],
    output: Union[str, pathlib.Path],
    chunk_size: int = 1024 * 1024 * 100,
) -> pathlib.Path:
    """Create a .mls file

    Args:
        files (List[str]): The list of files to be included in the
            MLSTAC file.
        output (Union[str, pathlib.Path]): The output file.

    Returns:
        pathlib.Path: The output file.
    """

    # Create the output file
    nmagic = b"#y"  # (31011).to_bytes(2, "little")
    dict_bytes = dict()

    # Get the total size of the files
    bytes_counter = 0
    for file in files:

        # Get the size of the file
        file_path = pathlib.Path(file)
        file_size: int = file_path.stat().st_size

        # Update the counter
        bytes_counter += file_size

        # Save the file size
        dict_bytes[file_path.stem] = [bytes_counter - file_size, file_size]

    # Encode the dictionary
    dict_bytes_enc: bytes = json.dumps(dict_bytes).encode()

    # Create the final file
    with open(output, "wb") as f:

        # Write the magic number
        f.write(nmagic)

        # Write the length of the HEADER
        f.write(len(dict_bytes_enc).to_bytes(8, "little"))

        # Write the HEADER
        f.write(dict_bytes_enc)

        # Write the data by chunks
        for file in tqdm.tqdm(files, desc="Creating MLSTAC file"):
            with open(file, "rb") as g:
                while True:
                    chunk = g.read(chunk_size)
                    if not chunk:
                        break
                    f.write(chunk)

    return output



from typing import List, Optional

import pydantic


class TensorDimension(pydantic.BaseModel):
    """In the ML-STAC specification, the dimensions are from the
    same modalities. This extension defines the dimensions of the samples
    for each Sample.

    fields:
        axis (int): The axis of the dimension.
        description (Optional[str]): A description of the dimension.
    """

    axis: int
    name: str
    description: Optional[str] = None


class TensorDimensions(pydantic.BaseModel):
    """In the ML-STAC specification, the dimensions are equal between
    all the samples. This extension defines the dimensions of the samples
    for each SampleTensor.

    fields:
        dimensions (Dict[str, Dimension]): A dictionary with the dimensions
            of the samples. The key is the name of the dimension and the
            value is a Dimension object.
        dtype (Optional[str]): The data type of the samples. This field
            is defined automatically by the 'automatic_field' method from the
            Collection class.
        shape (Optional[List[int]]): The shape of the samples. This field
            is defined automatically by the 'automatic_field' method from the
            Collection class.
        offsets (Optional[List[int]]): The offsets of the samples. This field
            is defined automatically by the 'automatic_field' method from the
            Collection class.
    """

    dimensions: Optional[List[TensorDimension]] = None



"""Core functions for reading and writing MLSTAC files.

This module contains the core functions for reading and writing MLS files.

# Local reading functions

- read_mlstac_metadata_local: Read the metadata of a MLSTAC file given a path.
- read_mlstac_data_local: Read the data of a MLSTAC file given a path.

# Remote reading functions

- read_mlstac_metadata_url: Read the metadata of a MLSTAC file given a URL.
- read_mlstac_data_url: Read the data of a MLSTAC file given a URL.

# Target encoding/decoding functions

- encode_json_array: Encode a dictionary as a NumPy array.
- decode_json_array: Decode a NumPy array as a dictionary.

# High-level functions
- load_metadata: Load the metadata of a MLSTAC file.
- load_data: Load the data of a MLSTAC file.
"""

import json
import pathlib
import struct
from io import BytesIO
from typing import List, Optional, Tuple, Union, Literal

import numpy as np
import pandas as pd
import rasterio as rio
import requests

import mlstac.utils

def read_mlstac_metadata_local(file: Union[str, pathlib.Path]) -> dict:
    """Read the metadata of a .mls file

    Args:
        file (Union[str, pathlib.Path]): A .mls file.

    Returns:
        dict: The metadata of the .mls file.
    """

    with open(file, "rb") as f:
        # Read the FIRST 2 bytes of the file
        nmagic = f.read(2)

        if nmagic != b"#y":
            raise ValueError("The file is not a '.mlstac' file")

        # Read the NEXT 8 bytes of the file
        header_len = f.read(8)

        # Convert bytes to uint64
        length = int.from_bytes(header_len, "little")

        # Read the HEADER considering the length (JSON)
        header = f.read(length)

        # Read the bytes 100-200
        header: dict = json.loads(header.decode())

    return header


def read_mlstac_data_local(
    file: Union[str, pathlib.Path],
    datapoint: pd.Series,
    metadata_length: Optional[dict] = None,
    backend: Literal["rasterio", "bytes"] = "rasterio"
) -> Tuple[np.ndarray, dict]:
    """Read the data of a mlstac file

    Args:
        file (Union[str, pathlib.Path]): The file to read.
        datapoint (pd.Series): The datapoint to read.
        metadata_length (Optional[dict], optional): The length of the
            metadata. If None, the code seeks the metadata of the file.
            Defaults to None.

    Returns:
        Tuple[np.ndarray, dict]: The data and metadata of the
            datapoint.
    """

    # Read the metadata
    if metadata_length is None:
        metadata = read_mlstac_metadata_local(file)
        metadata_length = len(json.dumps(metadata))

    # Define fseek
    fseek: int = datapoint["begin"] + metadata_length + 2 + 8

    with open(file, "rb") as f:
        # Move the pointer to the right position
        f.seek(fseek)

        # Read the specific bytes
        data: bytes = f.read(datapoint["length"])

        # Convert the bytes to a rasterio object
        with BytesIO() as bio:
            bio.write(data)
            bio.seek(0)
            if backend == "rasterio":
                with rio.open(bio) as src:
                    data = src.read()
                    metadata = src.meta
            else:
                data = bio.read()
                metadata = {}        
    return data, metadata


def read_mlstac_metadata_url(url: str) -> dict:
    """Read the metadata of a mlstac file given a URL. The
    server must support the HTTPS Range Request.

    Args:
        url (str): The URL of the file.

    Returns:
        Tuple[dict, int]: The metadata of the mlstac file
        and the length of the header.
    """

    # Fetch the first 8 bytes of the file
    headers = {"Range": "bytes=2-9"}
    response: requests.Response = requests.get(url, headers=headers)

    # Interpret the bytes as a little-endian unsigned 64-bit integer
    length_of_header: int = struct.unpack("<Q", response.content)[0] + 9

    # Fetch length_of_header bytes starting from the 9th byte
    headers = {"Range": f"bytes=10-{length_of_header}"}

    with requests.get(url, headers=headers) as response:
        # Interpret the response as a JSON object
        header = json.loads(response.content)

    return header


def read_mlstac_data_url(
    url: str, datapoint: pd.Series, metadata_length: Optional[int] = None, backend: Literal["rasterio", "bytes"] = "rasterio"
) -> Tuple[np.ndarray, dict]:
    """Read the data of a mlstac file given a URL.
    The server must support HTTPS Range Request.

    Args:
        URL (Union[str, pathlib.Path]): The file to be read.
        datapoint_id (str): The ID of the datapoint to read.
        length_metadata (Optional[int], optional): The length of the
            metadata. If None, the code seeks the metadata of the file.
            Defaults to None.

    Returns:
        Tuple[np.ndarray, dict]: The data and metadata of the
            datapoint.
    """

    # Read the metadata
    if metadata_length is None:
        metadata = read_mlstac_metadata_url(url)
        metadata_length = len(json.dumps(metadata))

    # Define the byte range to read the data
    fseek = datapoint["begin"] + metadata_length + 2 + 8

    # Define the HTTP Range Request
    headers = {"Range": f'bytes={fseek}-{fseek + datapoint["length"] - 1}'}

    # Fetch the data
    with requests.get(url, headers=headers) as response:
        response.raise_for_status()
        with BytesIO() as bio:
            bio.write(response.content)
            bio.seek(0)
            if backend == "rasterio":
                with rio.open(bio) as src:
                    data = src.read()
                    metadata = src.meta
            else:
                data = bio.read()
                metadata = {}

    return data, metadata

def encode_json_array(
    data: dict, shape: tuple[int, int], dtype: np.dtype
) -> np.ndarray:
    """Encode a dictionary as a NumPy array.

    Args:
        data (dict): The dictionary to encode.
        shape (tuple[int, int]): The shape of the NumPy array.
        dtype (np.dtype): The data type of the NumPy array.

    Returns:
        np.ndarray: The NumPy array.
    """

    # Convert dictionary to JSON string
    json_bytes = json.dumps(data).encode("utf-8")

    # Convert the NumPy array to bytes
    array = np.zeros(shape, dtype=dtype)
    byte_data = array.tobytes()

    # Convert from bytes to bytearray
    byte_data_arr = bytearray(byte_data)

    # Replace eight bytes with the length of the JSON string (UInt64)
    len_bytes = struct.pack("Q", len(json_bytes))
    byte_data_arr[:8] = len_bytes
    byte_data_arr[8 : (8 + len(json_bytes))] = json_bytes

    return np.frombuffer(byte_data_arr, dtype=np.uint16).reshape(array.shape)


def decode_json_array(array: np.ndarray) -> dict:
    """Decode a NumPy array as a dictionary.

    Args:
        data (dict): The dictionary to encode.
        shape (tuple[int, int]): The shape of the NumPy array.
        dtype (np.dtype): The data type of the NumPy array.

    Returns:
        np.ndarray: The NumPy array.
    """

    # Convert the NumPy array to bytes
    byte_data = array.tobytes()

    # Read the length of the JSON string
    len_bytes = byte_data[:8]
    header_len = struct.unpack("Q", len_bytes)[0]

    # Read the JSON string
    json_bytes = byte_data[8 : (8 + header_len)]

    # Convert the JSON string to a dictionary
    return json.loads(json_bytes.decode("utf-8"))


def load_metadata(path: Union[str, pathlib.Path]) -> pd.DataFrame:
    """Load a MLSTAC file

    Args:
        path (Union[str, pathlib.Path]): The path to the MLSTAC file.
            It can be a local path or a URL. If it is a URL, the
            server must support the HTTP Range requests.

    Returns:
        pd.DataFrame: The dataframe with three columns: 'datapoint_id',
            'begin', and 'length'. The 'datapoint_id' is the ID of the
            datapoint, 'begin' is the byte where the datapoint starts,
            and 'length' is the length of the datapoint in bytes.
    """

    # Convert the path to a string
    path = path.as_posix() if isinstance(path, pathlib.Path) else path

    # Obtain the file metadata
    if mlstac.utils.is_valid_url(path):
        dataset = read_mlstac_metadata_url(path)
        status = "remote"
    else:
        dataset = read_mlstac_metadata_local(path)
        status = "local"

    # Obtaint the length header
    length_dataset = len(json.dumps(dataset))

    # Convert dataset to DataFrame
    datapoints = list(dataset.items())
    metadata = pd.DataFrame(datapoints, columns=["datapoint_id", "values"])

    # Expand 'values' into 'begin' and 'length' columns
    metadata[["begin", "length"]] = pd.DataFrame(
        metadata["values"].tolist(), index=metadata.index
    )
    metadata = metadata.drop(columns="values")
    metadata.attrs.update(
        {"status": status, "path": path, "length_dataset": length_dataset}
    )

    return metadata


def load_data(
    dataset: Union[pd.DataFrame, pd.Series],
    save_metadata_datapoint: bool = False,
    return_generator: bool = False,
    backend: Literal["rasterio", "bytes"] = "rasterio",
    quiet: bool = False,
) -> Union[np.ndarray, List[Tuple[np.ndarray, dict]]]:
    """Download the data of a MLSTAC file.

    Args:
        dataset (pd.DataFrame): A DataFrame with byte ranges of
            the datapoints.
        save_metadata_datapoint (bool, optional): Each datapoint
            has associated metadata. If True, the function returns
            a list of tuples with the data and metadata of each
            datapoint. If False, the function returns a numpy array
            with the data of each datapoint. Defaults to False.
        return_generator (bool, optional): Whether to return a
            generator with the data of the datapoints. Defaults
            to False.
        quiet (bool, optional): Whether to print the progress of
            the download. If the file is local, the progress is
            not printed. Defaults to False.

    Returns:
        Union[np.ndarray, List[Tuple[np.ndarray, dict]]]: The data
            of the datapoints. If save_metadata_datapoint is True,
            the function returns a list of tuples with the data and
            metadata of each datapoint. If save_metadata_datapoint
            is False, the function returns a numpy array with the
            data of each datapoint.
    """

    if isinstance(dataset, pd.Series):
        dataset = pd.DataFrame(dataset).T

    # The dataset must have the attribute 'state' that indicates if the
    # data is remote or local.
    state: Literal["remote", "local"] = dataset.attrs["status"]


    def data_generator():
        for idx, row in dataset.iterrows():
            # Print the progress of the download
            if not quiet:
                print(f"Downloading datapoint: {row['datapoint_id']}")
            
            # Read the data and metadata of the datapoint according to the state
            if state == "remote":
                data, metadata = read_mlstac_data_url(
                    url=dataset.attrs["path"],
                    datapoint=row,
                    metadata_length=dataset.attrs["length_dataset"],
                    backend=backend
                )
            elif state == "local":
                data, metadata = read_mlstac_data_local(
                    file=dataset.attrs["path"],
                    datapoint=row,
                    metadata_length=dataset.attrs["length_dataset"],
                    backend=backend
                )

            # Whether to save the metadata of the datapoint
            if save_metadata_datapoint:
                yield (data, metadata)
            else:
                yield data

    # Return the data as a generator
    generator = data_generator()

    if return_generator:
        return generator
    else:
        data_list = list(generator)
        if save_metadata_datapoint:
            return data_list
        else:
            if backend == "rasterio":
                return np.array(data_list)
            return data_list


import datetime
import re
import warnings
from typing import Any, List, Tuple, Union

import numpy as np
import pydantic

SixFloats = Tuple[float, float, float, float, float, float]
CubeTime = Union[str, datetime.datetime, List[str], List[datetime.datetime]]


class Sample(pydantic.BaseModel):
    tensor: Any
    geotransform: SixFloats
    start_datetime: CubeTime
    crs: str
    end_datetime: Union[CubeTime, None] = None

    @pydantic.field_validator("tensor")
    def check_tensor(cls, v: np.ndarray) -> np.ndarray:
        if not isinstance(v, np.ndarray):
            raise ValueError("tensor must be a numpy array")
        return v

    @pydantic.field_validator("crs")
    def check_crs(cls, v: str) -> str:
        regex_exp = re.compile(r"^(?:EPSG|ESRI|SR-ORG):[0-9]+$")
        if not regex_exp.match(v):
            warnings.warn(
                "It is recommended crs to be from the format: <authority>:<code>"
            )
        return v

    @pydantic.model_validator(mode="after")
    def check_start_end_datetime(self) -> "Sample":
        # If the start_datetime is a list, the end_datetime must be a list
        # both lists must have the same length
        if isinstance(self.start_datetime, list):
            if self.end_datetime is not None:
                if not isinstance(self.end_datetime, list):
                    raise ValueError("end_datetime must be a list")
                if len(self.start_datetime) != len(self.end_datetime):
                    raise ValueError(
                        "start_datetime and end_datetime must have the same length"
                    )
            # the end_datetime must be always higher than the start_datetime
            for i in range(len(self.start_datetime)):
                if self.start_datetime[i] < self.end_datetime[i]:
                    raise ValueError("end_datetime must be after start_datetime")

        # If the start_datetime is a string, the end_datetime must be a string
        if isinstance(self.start_datetime, str):
            if self.end_datetime is not None:
                if not isinstance(self.end_datetime, str):
                    raise ValueError("end_datetime must be a string")

            # the end_datetime must be always higher than the start_datetime
            if self.start_datetime > self.end_datetime:
                raise ValueError("end_datetime must be after start_datetime")

        return self


## Create model squema
# demo = Sample(
#    tensor=np.random.rand(10, 10, 10),
#    geotransform=(0, 1, 0, 0, 0, 1),
#    start_datetime="2021-01-01",
#    end_datetime="2021-01-02",
#    crs="EPSG:4326",
# )
#
# import json
# with open("mlstac/specification/sample/squema.json", "w") as f:
#    json.dump(demo.model_json_schema(), f, indent=2)



from typing import List, Optional

import pydantic


class SpectralBand(pydantic.BaseModel):
    """This extension provides a way to define the spectral bands of a
    dataset. Useful for Remote Sensing datasets.

    fields:
        band (str): The name of the band.
        index (Optional[int]): The index of the band.
        description (Optional[str]): A description of the band.
        unit (Optional[str]): The unit of the band.
        wavelengths (Optional[List[float]]): The wavelengths of the band.
            It must be a list of two floats. The first float is the minimum
            wavelength and the second float is the maximum wavelength.
    """

    name: str
    index: Optional[int]
    common_name: Optional[str] = None
    description: Optional[str] = None
    unit: Optional[str] = None
    center_wavelength: Optional[float] = None
    full_width_half_max: Optional[float] = None


class SpectralBands(pydantic.BaseModel):
    """This extension provides a way to define the
    spectral bands of a dataset. Useful for Remote
    Sensing datasets.

    fields:
        bands (Dict[str, SpectralBand]): A dictionary with
            the spectral bands of the dataset. The key is the
            name of the band and the value is a SpectralBand
            object.
        axis (Optional[int]): The axis of the spectral bands.
        sensor (Optional[str]): The sensor of the spectral bands.
    """

    bands: List[SpectralBand]
    axis: Optional[int] = None
    sensor: Optional[str] = None



import re
from typing import List, Optional

import pydantic

from mlstac.specification.collection.stac import Link


class Info(pydantic.BaseModel):
    """Contact extension is part of the STAC extension.
    Info is a part of the Contact extension.
    """

    value: str
    roles: Optional[List[str]] = None

    @pydantic.field_validator("value")
    def check_value(cls, v):
        if not re.match(r"^\+[1-9]{1}[0-9]{3,14}$", v):
            if not re.match(r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$", v):
                raise ValueError("value must be a valid email or phone number")
        return v


class Address(pydantic.BaseModel):
    """Contact extension is part of the STAC extension.
    Address is a part of the Contact extension.
    """

    deliveryPoint: Optional[List[str]] = None
    city: Optional[str] = None
    administrativeArea: Optional[str] = None
    postalCode: Optional[str] = None
    country: Optional[str] = None


class Contact(pydantic.BaseModel):
    """Pydantic Contact extension from the STAC extension.
    More info: https://github.com/stac-extensions/contacts
    """

    name: str
    organization: str
    identifier: Optional[str] = None
    position: Optional[str] = None
    logo: Optional[Link] = None
    phones: Optional[List[Info]] = None
    emails: Optional[List[Info]] = None
    addresses: Optional[List[Address]] = None
    links: Optional[List[Link]] = None
    contactInstructions: Optional[str] = None
    roles: Optional[List[str]] = None



import re
from typing import List, Literal, Optional

import pydantic


class SingleCatalog(pydantic.BaseModel):
    name: Literal["train", "test", "validation"]
    data_files: List[pydantic.AnyUrl]
    data_descriptions: List[str]
    data_checksum: List[int]
    metadata_file: pydantic.AnyUrl
    metadata_description: str
    metadata_checksum: int

    @pydantic.field_validator("name")
    def name_must_be_valid(cls, value: str) -> str:
        if not re.match(r"train|test|val|validation", value):
            raise ValueError("name must be one of train, test, or validation.")
        return value


class Catalog(pydantic.BaseModel):
    train: SingleCatalog
    validation: Optional[SingleCatalog] = None
    test: Optional[SingleCatalog] = None



import gc
import json
import pathlib
import pickle
import tempfile
from typing import List, Literal, Tuple, Union

import numpy as np
import pandas as pd

import mlstac.datamodel
import mlstac.utils


def download(
    snippet: str,
    path: Union[str, pathlib.Path],
    split: Literal["train", "validation", "test", "all"] = "all",
    quiet: bool = False,
) -> pd.DataFrame:
    """Download the files of a MLSTAC collection.

    This function downloads the files of a MLSTAC collection. The
    function downloads the files of the collection according to the
    split. The split can be 'train', 'validation', 'test', or 'all'.

    Args:
        snippet (str): The URL of the MLSTAC Collection file.
        path (Union[str, pathlib.Path]): The path where the files
            will be downloaded.
        split: Literal["train", "validation", "test", "all"]: The
            split folder to download. The split can be 'train',
            'validation', 'test', or 'all'. Defaults to 'all'.
        quiet (bool, optional): Whether to print the progress of
            the download.

    Raises:
        ValueError: The split must be one of 'train', 'validation',
            'test', or 'all'.

    Returns:
        pd.DataFrame: The DataFrame with the files downloaded. There
            are three columns: 'url', 'local', and 'checksum'.
    """

    # Load the ML-STAC collection
    mlstac_collection, _ = mlstac.utils.load_mlstac_collection_file(snippet)
    mlstac_file = pathlib.Path(path) / "main.json"

    # Download the MLSTAC file
    with open(mlstac_file, "wb") as f:
        f.write(json.dumps(mlstac_collection).encode())

    # Prepare the files to download
    splits = {
        "all": ["train", "validation", "test"],
        "train": ["train"],
        "validation": ["validation"],
        "test": ["test"],
    }.get(split, None)

    if splits is None:
        raise ValueError(
            "The split must be one of 'train', 'validation', 'test', or 'all'."
        )

    # Define all the files to download and set it up the folder structures
    download_db: pd.DataFrame = mlstac.utils.from_mlstac_to_df_download(
        metadata=mlstac_collection, path=path, splits=splits
    )

    # Start the download
    for idx, row in download_db.iterrows():
        if not quiet:
            print(f"Downloading file {idx+1} of {len(download_db)}: {row['url']}")

        mlstac.utils.download_file_with_recovery(
            url=row["url"],
            local_filename=row["local"],
            checksum=row["checksum"],
            quiet=quiet,
        )

    return download_db


def load(snippet: Union[str, pathlib.Path, dict], force: bool = False) -> pd.DataFrame:
    """Load a MLSTAC collection.

    Args:
        snippet (Union[str, pathlib.Path, dict]): The URL of the MLSTAC
            Collection file, the path to the MLSTAC Collection file, or
            the MLSTAC Collection file as a dictionary.
        force (bool, optional): In the first time, the function saves
            the metadata as a parquet file in the temp folder. If the
            metadata file exists, the function loads the metadata from
            the file. To desactivate this behavior, set force to True.

    Returns:
        pd.DataFrame: The DataFrame with the MLSTAC Collection metadata.
    """

    # Before to prepare the metadata, check if the snippet is a dictionary
    tmp_folder = tempfile.gettempdir()

    if not isinstance(snippet, dict):
        tmp_filename = mlstac.utils.compress_string(str(snippet))
        tmp_mtd = pathlib.Path(tmp_folder) / tmp_filename
        condition = tmp_mtd.exists() and not force
    else:
        tmp_filename = mlstac.utils.compress_string(snippet["id"])
        tmp_mtd = pathlib.Path(tmp_folder) / tmp_filename
        condition = False

    # If the metadata exists, load the metadata from pickle file
    if condition:
        with open(tmp_mtd, "rb") as f:
            mlstac_object = pickle.load(f)
    else:

        # Load the ML-STAC collection
        mlstac_collection, state = mlstac.utils.load_mlstac_collection_file(
            snippet=snippet
        )

        # Prepare the metadata
        metadata_total = mlstac.utils.from_mlstac_to_metadata_parquet(
            mlstac_collection=mlstac_collection, state=state, snippet=snippet
        )

        # Return a MLSTAC object
        mlstac_object = mlstac.datamodel.MLSTAC(
            metadata=metadata_total, collection=mlstac_collection
        )

        # Save the MLS
        with open(tmp_mtd, "wb") as f:
            pickle.dump(mlstac_object, f)

    return mlstac_object


def get_data(
    dataset: Union[pd.DataFrame, pd.Series],
    save_metadata_datapoint: bool = False,
    return_generator: bool = False,
    quiet: bool = False,
) -> Union[np.ndarray, List[Tuple[np.ndarray, dict]]]:
    """Download the data of a MLSTAC file.

    Args:
        dataset (pd.DataFrame): A DataFrame with byte ranges of
            the datapoints.
        save_metadata_datapoint (bool, optional): Each datapoint
            has associated metadata. If True, the function returns
            a list of tuples with the data and metadata of each
            datapoint. If False, the function returns a numpy array
            with the data of each datapoint. Defaults to False.
        return_generator (bool, optional): Whether to return a
            generator with the data of the datapoints. Defaults
            to False.
        quiet (bool, optional): Whether to print the progress of
            the download. If the file is local, the progress is
            not printed. Defaults to False.

    Returns:
        Union[np.ndarray, List[Tuple[np.ndarray, dict]]]: The data
            of the datapoints. If save_metadata_datapoint is True,
            the function returns a list of tuples with the data and
            metadata of each datapoint. If save_metadata_datapoint
            is False, the function returns a numpy array with the
            data of each datapoint.
    """

    if isinstance(dataset, pd.Series):
        dataset = pd.DataFrame(dataset).T

    # The dataset must have the column 'state' that indicates if the
    # data is remote or local
    state = dataset.iloc[0]["state"]

    def data_generator():
        for idx, row in dataset.iterrows():
            # Print the progress of the download
            if not quiet:
                print(f"Reading datapoint: {row['datapoint_id']}")

            if state == "remote":
                # Read the data and metadata of the datapoint
                data, metadata = mlstac.core.read_mlstac_data_url(
                    url=row.url,
                    datapoint=row,
                    metadata_length=row.metadata_length,
                )
            else:
                # Read the data and metadata of the datapoint
                data, metadata = mlstac.core.read_mlstac_data_local(
                    file=row.file_path, datapoint=row, metadata_length=row.metadata_length
                )

            # Whether to save the metadata of the datapoint
            if save_metadata_datapoint:
                yield (data, metadata)
            else:
                yield data

    # Return the data as a generator
    generator = data_generator() 

    if return_generator:
        return generator
    else:
        data_list = list(generator)
        if save_metadata_datapoint:
            return data_list
        else:
            return np.array(data_list)


def prepare_data(
    mls_path: Union[str, pathlib.Path],
    safetensor_path: Union[str, pathlib.Path],
    split: Literal["train", "val", "validation", "test", "all"] = "all",
    verbose: bool = False,
):
    pass



""" This module contains the STAC specification version 1.0.0 implemented 
as pydantic models. The STAC specification documentation can be found at
https://github.com/radiantearth/stac-spec/blob/master/collection-spec/collection-spec.md.
The optional fields 'stac_version', 'summaries' and 'assets' are not implemented 
as they are not required for the ML-STAC specification.  The fields are 
organized into required and optional fields.  The required fields are 
those that are required by the STAC specification in order to create a 
valid STAC Collection.
"""

import re
from typing import Any, Dict, List, Literal, Optional

import pydantic

# Required fields -------------------------------------------------------------


class Type(pydantic.BaseModel):
    """REQUIRED. Must be set to Collection to be a valid Collection."""

    field: Literal["Collection"]


class Version(pydantic.BaseModel):
    """REQUIRED. In the ML-STAC specification, the STAC version is
    fixed to 1.0.0.

    Fields:
        stac_version (Literal["1.0.0"]): The version of the STAC
        specification.
    """

    field: Literal["1.0.0"]


class Id(pydantic.BaseModel):
    """REQUIRED. It is important that Collection identifiers
    are unique across the provider. And providers should strive as much as
    possible to make their Collection ids 'globally' unique, prefixing any
    common information with a unique string. This could be the provider's
    name if it is a fairly unique name, or their name combined with the domain
    they operate in.
    """

    field: str

    @pydantic.field_validator("field")
    def check_id(cls, v):
        if len(v) < 1:
            raise ValueError("Must be at least 1 characters")
        return v


class Description(pydantic.BaseModel):
    """REQUIRED. Detailed multi-line description to fully explain the Collection.
    CommonMark 0.29 syntax MAY be used for rich text representation.
    """

    field: str

    @pydantic.field_validator("field")
    def check_id(cls, v):
        if len(v) < 1:
            raise ValueError("Must be at least 1 characters")
        return v


class License(pydantic.BaseModel):
    """REQUIRED. Collection's license(s), either a SPDX License identifier, various
    if multiple licenses apply or proprietary for all other cases.
    """

    field: str

    @pydantic.field_validator("field")
    def check_license(cls, v):
        regex_exp = re.compile("^[\\w\\-\\.\\+]+$")
        if not re.match(regex_exp, v):
            raise ValueError("Must be a valid SPDX License identifier")
        return v


class SpatialExtent(pydantic.BaseModel):
    """REQUIRED. The object describes the spatial extents of the Collection.
    bbox: Each outer array element can be a separate spatial extent describing
    the bounding boxes of the assets represented by this Collection using either
    2D or 3D geometries.

    The first bounding box always describes the overall spatial extent of the data.
    All subsequent bounding boxes can be used to provide a more precise description
    of the extent and identify clusters of data. Clients only interested in the overall
    spatial extent will only need to access the first item in each array. It is recommended
    to only use multiple bounding boxes if a union of them would then include a large uncovered
    area (e.g. the union of Germany and Chile).

    The length of the inner array must be 2*n where n is the number of dimensions. The array
    contains all axes of the southwesterly most extent followed by all axes of the
    northeasterly most extent specified in Longitude/Latitude or Longitude/Latitude/Elevation
    based on WGS 84. When using 3D geometries, the elevation of the southwesterly most extent
    is the minimum depth/height in meters and the elevation of the northeasterly most extent
    is the maximum.

    The coordinate reference system of the values is WGS 84 longitude/latitude. Example that
    covers the whole Earth: [[-180.0, -90.0, 180.0, 90.0]]. Example that covers the whole
    earth with a depth of 100 meters to a height of 150 meters: [[-180.0, -90.0, -100.0,
    180.0, 90.0, 150.0]].
    """

    bbox: List[List[float]]

    @pydantic.field_validator("bbox")
    def check_spatial_extent(cls, v):
        for item in v:
            for value in item:
                if value < -180 or value > 180:
                    raise ValueError("Longitude must be between -180 and 180")
        return v

    @pydantic.field_validator("bbox")
    def check_spatial_length(cls, v):
        if len(v) < 1:
            raise ValueError("Must be at least 1 spatial extent")
        return v

    @pydantic.field_validator("bbox")
    def check_spatial_dimension(cls, v):
        for item in v:
            if not (len(item) == 4 or len(item) == 6):
                raise ValueError("Must be 2D or 3D geometries")
        return v


class TemporalExtent(pydantic.BaseModel):
    """REQUIRED. The object describes the temporal extents of the Collection. interval: Each outer array
    element can be a separate temporal extent. The first time interval always describes the
    overall temporal extent of the data. All subsequent time intervals can be used to provide a
    more precise description of the extent and identify clusters of data. Clients only interested
    in the overall extent will only need to access the first item in each array. It is recommended
    to only use multiple temporal extents if a union of them would then include a large uncovered
    time span (e.g. only having data for the years 2000, 2010 and 2020).

    Each inner array consists of exactly two elements, either a timestamp or null.

    Timestamps consist of a date and time in UTC and MUST be formatted according to RFC 3339,
    section 5.6. The temporal reference system is the Gregorian calendar.

    Open date ranges are supported by setting the start and/or the end time to null. Example
    for data from the beginning of 2019 until now: [["2019-01-01T00:00:00Z", null]]. It is recommended
    to provide at least a rough guideline on the temporal extent and thus it's not recommended to
    set both start and end time to null. Nevertheless, this is possible if there's a strong use case
    for an open date range to both sides.
    """

    interval: List[List[Optional[str]]]

    @pydantic.field_validator("interval")
    def check_temporal_extent(cls, v):
        if v[0][0] is None and v[0][1] is None:
            raise ValueError("Both start and end time cannot be None")
        return v

    @pydantic.field_validator("interval")
    def check_temporal_length(cls, v):
        if len(v) < 1:
            raise ValueError("Must be at least 1 temporal extent")
        return v

    @pydantic.field_validator("interval")
    def check_temporal_dimension(cls, v):
        for item in v:
            if not (len(item) == 2):
                raise ValueError("Must be a length of 2")
        return v

    @pydantic.field_validator("interval")
    def check_temporal_regex(cls, v):
        regex_exp = re.compile("(\\+00:00|Z)$")
        for item in v:
            for value in item:
                if value is not None:
                    if not regex_exp.search(value):
                        raise ValueError("Must be a valid RFC 3339 timestamp")
        return v


class Extent(pydantic.BaseModel):
    """REQUIRED. The object describes the spatio-temporal extents of the Collection.
    Both spatial and temporal extents are required to be specified.
    """

    spatial: SpatialExtent
    temporal: TemporalExtent


class Link(pydantic.BaseModel):
    """This object describes a relationship with another entity. Data
    providers are advised to be liberal with links. For a full discussion of the
    situations where relative and absolute links are recommended see the 'Use of links'
    section of the STAC best practices.
    https://github.com/radiantearth/stac-spec/blob/master/best-practices.md#use-of-links
    """

    href: str
    rel: str
    type: Optional[str] = None
    title: Optional[str] = None


class Links(pydantic.BaseModel):
    """REQUIRED. A list of links related to this Collection. The relationships
    of the links are defined by the link object with the relation type.
    """

    field: List[dict]


# Optional fields -------------------------------------------------------------
class Title(pydantic.BaseModel):
    """OPTIONAL: A short descriptive one-line title for the Collection."""

    field: str


class Keywords(pydantic.BaseModel):
    """OPTIONAL: A list of keywords describing the Collection."""

    field: List[str]


class Provider(pydantic.BaseModel):
    """A STAC required field.  A provider is any of the
    organizations that captures or processes the content
    of the Collection and therefore influences the data
    offered by this Collection. May also include information
    about the final storage provider hosting the data.

    Fields:
        name (str):  REQUIRED. The name of the organization or the individual
        description (str):  Multi-line description to add further provider
            information such as processing details for processors and
            producers, hosting details for hosts or basic contact information.
            CommonMark 0.29 syntax MAY be used for rich text representation.
        roles (List[Literal["licensor", "producer", "processor", "host"]]):
            Roles of the provider.
        url (str):  A URL to a web page with more information about the
            provider.
    """

    name: str
    description: Optional[str] = None
    roles: Optional[List[Literal["licensor", "producer", "processor", "host"]]] = None
    url: Optional[str] = None


class Providers(pydantic.BaseModel):
    """A STAC required field. A list of providers, which may include
    all organizations capturing or processing the data or the hosting
    provider. Providers should be listed in chronological order with
    the most recent provider being the last element of the list. See
    the STAC Provider Extension for more information.
    """

    field: List[Provider]


class Assets(pydantic.BaseModel):
    """OPTIONAL. A dictionary of asset objects that can
    be downloaded.
    """

    href: str
    title: Optional[str] = None
    description: Optional[str] = None
    type: Optional[str] = None
    roles: Optional[List[str]] = None


class ItemAsset(pydantic.BaseModel):
    """OPTIONAL. A dictionary of asset objects that can
    be downloaded.
    """

    title: Optional[str] = None
    description: Optional[str] = None
    type: Optional[str] = None
    roles: Optional[List[str]] = None


class STACCollection(pydantic.BaseModel):
    """A STAC required field. The object describes the spatio-temporal
    extents of the Collection. Both spatial and temporal extents are
    required to be specified.
    """

    type: str = Type(field="Collection").field
    stac_version: str = Version(field="1.0.0").field
    stac_extensions: Optional[List[str]] = None

    id: str
    description: str
    license: str
    extent: Extent
    links: Optional[List[Link]] = [
        Link(
            rel="self",
            href="collection.json",
            type="application/json",
            title="An ML-STAC Collection JSON file",
        )
    ]
    item_assets: Optional[Dict[str, ItemAsset]] = {
        "collection": ItemAsset(
            title="An ML-STAC Item JSON file",
            type="application/json",
            roles=["mlstac-collection"],
        ),
        "data": ItemAsset(
            title="A collection of .mls files",
            type="application/mls; profile=cloud-optimized",
            roles=["data"],
        ),
        "metadata": ItemAsset(
            title="A collection of .parquet files",
            type="application/parquet",
            roles=["metadata"],
        ),
    }

    # Optional STAC fields -------------------------------
    title: Optional[str] = None
    keywords: Optional[List[str]] = None
    providers: Optional[Providers] = None
    summaries: Optional[Any] = None
    assets: Optional[Dict[str, Assets]] = None
